import json
import os
import subprocess
from pathlib import Path
import logging
from go4py.types import GoFunction


logger = logging.getLogger(__name__)


def get_go_functions(module_folder: str):
    """list all go exported functions in the go module"""

    # Path to the generated functions.json file
    functions_json_path = Path("artifacts/functions.json")

    # Check if the functions.json file was generated
    if not functions_json_path.exists():
        logger.error(f"Error: {functions_json_path} was not generated")
        raise FileNotFoundError(f"{functions_json_path} was not generated")

    # Read the functions.json file
    logger.debug(f"Reading functions from: {functions_json_path.name}")
    with open(functions_json_path, "r") as f:
        functions_data = json.load(f)

    # Generate GoFunction objects from the JSON data
    go_functions = []
    for func_data in functions_data:
        try:
            go_function = GoFunction.model_validate(func_data)
            go_functions.append(go_function)
            # print(f"Parsed function: {go_function.name}")
        except Exception as e:
            logger.warning(
                f"function skipped: {func_data['name']} (set log-level to DEBUG for more info)"
            )
            logger.debug(f"Error: {e}")

    logger.info(f"Successfully parsed {len(go_functions)} functions")

    # Return the list of GoFunction objects for further processing
    return go_functions



from rich.console import Console
from rich.panel import Panel

console = Console()


def print_text_box(text, bold_text=""):
    if bold_text:
        text += f"  [bold]{bold_text}[/bold] "
    panel = Panel(f" {text} ", title_align="center", border_style="green", expand=False)
    console.print(panel)


if __name__ == "__main__":
    print_text_box("hi", "wed")

"""
uv run -m go4py.utils.text_util
"""



from go4py.code_gen.slice import go_slice_from_py_list
from go4py.types import ByteSliceType, GoStringType, SliceType, UnknownType, Variable


def gen_go_copy(v: Variable, free_logic: str):
    """
    this function will convert the c type to go type variable
    it will also prefixed the name with `go_`

    """
    match v.type:
        case GoStringType():
            copy_logic = f"""
    GoString go_{v.name} = {{{v.name}, (GoInt)strlen({v.name})}};"""
            return copy_logic, free_logic
        case SliceType():
            if type(v.type.item_type) is UnknownType:
                raise Exception("Nested types are not supported! (except for [][]byte)")
            return go_slice_from_py_list(v, free_logic)
        case ByteSliceType():
            copy_logic = f"""
    GoInt len_{v.name} = PyBytes_Size({v.name});
    GoSlice go_{v.name} = {{PyBytes_AsString({v.name}), len_{v.name}, len_{v.name}}};"""
            return copy_logic, free_logic
        case _:
            raise NotImplementedError()



import re


def render_template(template_string, data):
    """
    A very basic template engine that replaces placeholders with values from a dictionary.

    Args:
      template_string: The string containing placeholders like {{ variable_name }}.
      data: A dictionary where keys are variable names and values are their corresponding values.

    Returns:
      The template string with placeholders replaced by their values.
      Returns the original template if a placeholder is not found in the data.
    """

    def replace_placeholder(match):
        variable_name = match.group(1).strip()
        try:
            return str(data[variable_name])
        except KeyError:
            return match.group(0)  # Return the original placeholder if not found

    return re.sub(r"{{(.*?)}}", replace_placeholder, template_string)



from go4py.code_gen.copy_logic import gen_go_copy
from go4py.code_gen.slice import indent
from go4py.doc_annotation import DocAnnots, make_doc_annots
from go4py.types import (
    BoolType,
    ByteSliceType,
    CStringType,
    FloatType,
    GoFunction,
    GoStringType,
    IntType,
    SliceType,
    UnknownType,
    VarType,
    Variable,
)


module_name = "go_cool"


class ArgumentParser:
    def __init__(self):
        self.arg_decl = []
        self.format_string = ""
        self.arg_list = []
        self.check_logic = ""
        self.copy_logic = ""
        self.free_logic = ""

    def addArg(self, var: Variable):
        self.arg_decl.append(f"    {var.type.c_type()} {var.name};")
        if var.type.need_copy:
            copy_code, free_logic = gen_go_copy(var, self.free_logic)
            self.copy_logic += copy_code
            self.free_logic = free_logic

        self.format_string += var.type.fmt_str()
        self.arg_list.append("&" + var.name)

    def gen_ParseTuple(self):
        if len(self.arg_decl) == 0:
            return ""
        args = "\n".join(self.arg_decl)
        return f"""\n{args}
    if (!PyArg_ParseTuple(args, "{self.format_string}", {", ".join(self.arg_list)}))
        return NULL;"""

    def gen_checks(self):
        # TODO
        return ""

    def gen_copys(self):
        return self.copy_logic

    def gen_code(self):
        result = self.gen_ParseTuple()
        result += self.gen_checks()
        result += self.gen_copys()
        return result


def get_return_c_type(fn: GoFunction) -> str:
    if len(fn.return_type) == 0:
        raise Exception("no return type")
    elif len(fn.return_type) == 1:
        return fn.return_type[0].cgo_type()
    else:
        return f"struct {fn.name}_return"


def gen_fn_call(fn: GoFunction, doc_annots: DocAnnots):
    args = [("go_" + a.name if a.type.need_copy else a.name) for a in fn.arguments]

    if len(fn.return_type) == 0:
        if doc_annots.no_gil:
            before = "    Py_BEGIN_ALLOW_THREADS\n"
            after = "\n    Py_END_ALLOW_THREADS"
        else:
            before = after = ""
        # call the function without return value
        return f"{before}    {fn.name}({','.join(args)});{after}"
    else:
        if doc_annots.no_gil:
            return f"""    {get_return_c_type(fn)} result;
    Py_BEGIN_ALLOW_THREADS
    result = {fn.name}({",".join(args)});
    Py_END_ALLOW_THREADS"""
        else:
            # call the function with return value
            return f"    {get_return_c_type(fn)} result = {fn.name}({','.join(args)});"


class ReturnConverter:
    def __init__(self, t: VarType, msgpack_decode: False, var="result"):
        self.t = t
        self.var = var
        self.py_name = "py_" + var.replace(".", "_")
        self.reduceable = type(t) in [IntType, FloatType, BoolType]
        self.is_msgpack = (type(t) is ByteSliceType) and msgpack_decode
        self.msgpack_name = self.py_name + "_msgpack" if self.is_msgpack else None

    def return_var(self):
        if self.reduceable:
            return self.t.converter(self.var)
        elif self.is_msgpack:
            return self.msgpack_name
        else:
            return self.py_name

    def nullable_var(self):
        T = type(self.t)
        if T in [SliceType, ByteSliceType]:
            return f"{self.var}.data"
        if T == GoStringType:
            return f"{self.var}.p"
        if T == CStringType:
            return self.var
        return None

    def gen_py_result(self):
        if type(self.t) is SliceType:
            item_t = self.t.item_type
            item_converter = ReturnConverter(item_t, False, "item")
            return f"""
    PyObject* {self.py_name};
    if ({self.nullable_var()} == NULL) {{
        {self.py_name} = GetPyNone();
    }} else {{
        {self.py_name} = PyList_New({self.var}.len);
        for (int i = 0; i < {self.var}.len; i++) {{
            {item_t.cgo_type()} item = (({item_t.cgo_type()}*){self.var}.data)[i];{indent(item_converter.gen_code(), 8)}
            PyList_SetItem({self.py_name}, i, {item_converter.return_var()});
        }}
    }}"""
        elif self.is_msgpack:
            return f"""
    PyObject* {self.msgpack_name};
    if ({self.var}.data!=NULL){{
        PyObject* {self.py_name} = PyBytes_FromStringAndSize({self.var}.data, {self.var}.len);
        {self.msgpack_name} = PyObject_CallFunctionObjArgs(unpackb, {self.py_name}, NULL);
        Py_DECREF({self.py_name});
    }}else{{
        {self.msgpack_name} = GetPyNone();
    }}"""

        else:
            return f"\n    PyObject* {self.py_name} = {self.nullable_var()}==NULL ? GetPyNone() : {self.t.converter(self.var)};"

    def gen_copys(self):
        return ""

    def gen_free_and_refdec(self):
        if self.t.need_free():
            if type(self.t) in [SliceType, ByteSliceType]:
                return f"\n    free({self.var}.data);"
            else:
                return f"\n    free({self.var});"
        else:
            return ""

    def gen_code(self):
        if self.reduceable:
            return ""
        result = self.gen_py_result()
        if self.t.need_copy:
            result += self.gen_copys()
        result += self.gen_free_and_refdec()
        return result


def gen_return_code(fn: GoFunction, doc_annots: DocAnnots) -> str:
    return_types = fn.return_type

    code = ""

    for i, t in enumerate(return_types):
        if type(t) is UnknownType:
            return_types[i] = t.resolve()
    if len(return_types) == 0:
        code += "\n    RETURN_NONE;"
    else:
        msgpack_decode = doc_annots.msgpack_decode
        if len(return_types) == 1:
            conv = ReturnConverter(return_types[0], msgpack_decode, "result")
            code += conv.gen_code() + f"\n    return {conv.return_var()};"
        else:
            return_converters = [
                ReturnConverter(t, msgpack_decode, f"result.r{i}")
                for i, t in enumerate(return_types)
            ]
            code = ""
            for c in return_converters:
                # breakpoint()
                if not c.reduceable:
                    code += c.gen_code()
            code += '\n    PyObject* py_result = Py_BuildValue("'
            for t in return_types:
                if c.reduceable:
                    code += t.fmt_str()
                else:
                    code += "O"
            code += '"'
            for c in return_converters:
                if c.reduceable:
                    code += f", {c.var}"
                else:
                    code += f", {c.return_var()}"
            code += ");"
            for c in return_converters:
                if not c.reduceable:
                    code += f"\n    Py_DECREF({c.return_var()});"

            code += "\n    return py_result;"
    return code


def gen_fn(fn: GoFunction, module_name: str) -> str:
    doc_annots = fn.doc_annots()

    if doc_annots.skip_binding:
        return f"\n// function {fn.name} is skipped due to 'skip-binding' annotation\n"

    arg_parser = ArgumentParser()
    for arg in fn.arguments:
        if type(arg.type) is UnknownType:
            arg.type = arg.type.resolve()
        arg_parser.addArg(arg)
    return_code = gen_return_code(fn, doc_annots)

    return f"""
static PyObject* {module_name}_{fn.lowercase_name()}(PyObject* self, PyObject* args) {{ {arg_parser.gen_code()}
{gen_fn_call(fn, doc_annots)}{arg_parser.free_logic}{return_code}
}}"""



from go4py.types import UnknownType, Variable, VarType, GoStringType

need_c_convert = [GoStringType]


def indent(str, indent=4):
    return "\n".join([(indent * " " + line) if line else "" for line in str.split("\n")])


class ItemConverter:
    def __init__(
        self,
        item_type: VarType,
        free_resource_code: str,
        item_name="item",
        # indent = 0,
    ):
        self.t = item_type
        if type(self.t) is UnknownType:
            self.t = self.t.resolve()
        self.name = item_name
        self.free_resource_code = indent(free_resource_code, 8)

    def item_c_type(self):
        c_type = self.t.c_type()
        if c_type == "char*":
            c_type = "const char*"
        return c_type

    def item_cgo_type(self):
        cgo_type = self.t.cgo_type()
        if cgo_type == "char*":
            cgo_type = "const char*"
        return cgo_type

    def check_and_convert(self):
        pytype = self.t.check("").split("_Check")[0]  # TODO: this is hacky fix it
        result = f"""if (!{self.t.check(self.name)}) {{
            PyErr_SetString(PyExc_TypeError, "List items must be {pytype}");{self.free_resource_code}
            return NULL;
        }}"""
        if self.t.need_copy:
            result += f"\n        {self.item_c_type()} c_{self.name} = {self.t.from_py_converter(self.name)};"
        return result

    def final_value(self):
        if self.t.need_copy:
            match self.t:
                case GoStringType():
                    return f"(GoString) {{c_{self.name}, (GoInt)strlen(c_{self.name})}}"
                case _:
                    raise Exception("Not implemented")
        else:
            return self.t.from_py_converter(self.name)


def go_slice_from_py_list(inp_var: Variable, other_free_code=""):
    name = inp_var.name
    free_logic = f"""{other_free_code}
    free({name}_CArray);"""
    item_conv = ItemConverter(inp_var.type.item_type, free_logic, "item")
    copy_logic = f"""
    if (!PyList_Check({name})) {{
        PyErr_SetString(PyExc_TypeError, "Argument {name} must be a list");{indent(other_free_code)}
        return NULL;
    }}
    int len_{name} = PyList_Size({name});
    {item_conv.item_cgo_type()}* {name}_CArray = malloc(len_{name} * sizeof({inp_var.type.item_type.cgo_type()}));
    for (int i = 0; i < len_{name}; i++) {{
        PyObject* item = PyList_GetItem({name}, i);
        {item_conv.check_and_convert()}
        {name}_CArray[i] = {item_conv.final_value()};
    }}
    if (PyErr_Occurred()) {{{indent(free_logic)}
        return NULL;
    }}
    GoSlice go_{name} = {{{name}_CArray, (GoInt)len_{name}, (GoInt)len_{name}}};"""
    return copy_logic, free_logic



import pytest


from go4py.types import IntType, CStringType, GoStringType, BoolType, FloatType, Variable

int_test_cases = {}

v = Variable(name="v", type={"go_type": "int"})
int_test_cases["int"] = (v, 64, False)
v = Variable(name="v", type={"go_type": "int8"})
int_test_cases["int8"] = (v, 8, False)
v = Variable(name="v", type={"go_type": "uint8"})
int_test_cases["uint8"] = (v, 8, True)
v = Variable(name="v", type={"go_type": "uint16"})
int_test_cases["uint16"] = (v, 16, True)
v = Variable(name="v", type={"go_type": "int32"})
int_test_cases["int32"] = (v, 32, False)
v = Variable(name="v", type={"go_type": "uint32"})
int_test_cases["uint32"] = (v, 32, True)
v = Variable(name="v", type={"go_type": "int64"})
int_test_cases["int64"] = (v, 64, False)
v = Variable(name="v", type={"go_type": "uint64"})
int_test_cases["uint64"] = (v, 64, True)
v = Variable(name="v", type={"go_type": "byte"})
int_test_cases["byte"] = (v, 8, True)
v = Variable(name="v", type={"go_type": "rune"})
int_test_cases["rune"] = (v, 32, False)


@pytest.mark.parametrize("var,bits,unsigned", int_test_cases.values(), ids=int_test_cases.keys())
def test_int_type(var, bits, unsigned):
    assert type(var.type) is IntType
    assert var.type.bits == bits
    assert var.type.unsigned == unsigned


float_test_cases = {}
v = Variable(name="v", type={"go_type": "float32"})
float_test_cases["float32"] = (v, 32)
v = Variable(name="v", type={"go_type": "float64"})
float_test_cases["float64"] = (v, 64)


@pytest.mark.parametrize("var,bits", float_test_cases.values(), ids=float_test_cases.keys())
def test_float_type(var, bits):
    assert type(var.type) is FloatType
    assert var.type.bits == bits


others_test_cases = {}
v = Variable(name="v", type={"go_type": "string"})
others_test_cases["string"] = (v, GoStringType)
v = Variable(name="v", type={"go_type": "*C.char"})
others_test_cases["c_char"] = (v, CStringType)
v = Variable(name="v", type={"go_type": "bool"})
others_test_cases["bool"] = (v, BoolType)


@pytest.mark.parametrize(
    "var,expected_type", others_test_cases.values(), ids=others_test_cases.keys()
)
def test_other_types(var, expected_type):
    assert type(var.type) is expected_type



import pytest
from go4py.code_gen.generate_wrapper import gen_fn
from go4py.types import GoFunction

test_cases = {}

fn = GoFunction(
    name="Func_1",
    arguments=[
        {"name": "a", "type": {"go_type": "[]int"}},
    ],
    return_type=[],
)
fn_res = """
static PyObject* test_func_1(PyObject* self, PyObject* args) { 
    PyObject* a;
    if (!PyArg_ParseTuple(args, "O", &a))
        return NULL;
    if (!PyList_Check(a)) {
        PyErr_SetString(PyExc_TypeError, "Argument a must be a list");
        return NULL;
    }
    int len_a = PyList_Size(a);
    long* a_CArray = malloc(len_a * sizeof(long));
    for (int i = 0; i < len_a; i++) {
        PyObject* item = PyList_GetItem(a, i);
        if (!PyLong_Check(item)) {
            PyErr_SetString(PyExc_TypeError, "List items must be PyLong");
            free(a_CArray);
            return NULL;
        }
        a_CArray[i] = PyLong_AsLong(item);
    }
    if (PyErr_Occurred()) {
        free(a_CArray);
        return NULL;
    }
    GoSlice go_a = {a_CArray, (GoInt)len_a, (GoInt)len_a};
    Func_1(go_a);
    free(a_CArray);
    RETURN_NONE;
}"""
test_cases["int_slice_input"] = (fn, fn_res)


fn = GoFunction(
    name="Func_2",
    arguments=[
        {"name": "a", "type": {"go_type": "[]string"}},
        {"name": "b", "type": {"go_type": "[]*C.char"}},
    ],
    return_type=[],
)
fn_res = """
static PyObject* test_func_2(PyObject* self, PyObject* args) { 
    PyObject* a;
    PyObject* b;
    if (!PyArg_ParseTuple(args, "OO", &a, &b))
        return NULL;
    if (!PyList_Check(a)) {
        PyErr_SetString(PyExc_TypeError, "Argument a must be a list");
        return NULL;
    }
    int len_a = PyList_Size(a);
    GoString* a_CArray = malloc(len_a * sizeof(GoString));
    for (int i = 0; i < len_a; i++) {
        PyObject* item = PyList_GetItem(a, i);
        if (!PyUnicode_Check(item)) {
            PyErr_SetString(PyExc_TypeError, "List items must be PyUnicode");
            free(a_CArray);
            return NULL;
        }
        const char* c_item = PyUnicode_AsUTF8(item);
        a_CArray[i] = (GoString) {c_item, (GoInt)strlen(c_item)};
    }
    if (PyErr_Occurred()) {
        free(a_CArray);
        return NULL;
    }
    GoSlice go_a = {a_CArray, (GoInt)len_a, (GoInt)len_a};
    if (!PyList_Check(b)) {
        PyErr_SetString(PyExc_TypeError, "Argument b must be a list");
        free(a_CArray);
        return NULL;
    }
    int len_b = PyList_Size(b);
    const char** b_CArray = malloc(len_b * sizeof(char*));
    for (int i = 0; i < len_b; i++) {
        PyObject* item = PyList_GetItem(b, i);
        if (!PyUnicode_Check(item)) {
            PyErr_SetString(PyExc_TypeError, "List items must be PyUnicode");
            free(a_CArray);
            free(b_CArray);
            return NULL;
        }
        b_CArray[i] = PyUnicode_AsUTF8(item);
    }
    if (PyErr_Occurred()) {
        free(a_CArray);
        free(b_CArray);
        return NULL;
    }
    GoSlice go_b = {b_CArray, (GoInt)len_b, (GoInt)len_b};
    Func_2(go_a,go_b);
    free(a_CArray);
    free(b_CArray);
    RETURN_NONE;
}"""
test_cases["string_slice_inputs"] = (fn, fn_res)
print(gen_fn(fn, "test"))

#####

fn = GoFunction(
    name="Func_3",
    arguments=[],
    return_type=[{"go_type": "[]int"}],
)
fn_res = """
static PyObject* test_func_3(PyObject* self, PyObject* args) { 
    GoSlice result = Func_3();
    PyObject* py_result;
    if (result.data == NULL) {
        py_result = GetPyNone();
    } else {
        py_result = PyList_New(result.len);
        for (int i = 0; i < result.len; i++) {
            long item = ((long*)result.data)[i];
            PyList_SetItem(py_result, i, PyLong_FromLong(item));
        }
    }
    free(result.data);
    return py_result;
}"""
test_cases["return_int_slice"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_4",
    arguments=[],
    return_type=[{"go_type": "[]*C.char"}],
)
fn_res = """
static PyObject* test_func_4(PyObject* self, PyObject* args) { 
    GoSlice result = Func_4();
    PyObject* py_result;
    if (result.data == NULL) {
        py_result = GetPyNone();
    } else {
        py_result = PyList_New(result.len);
        for (int i = 0; i < result.len; i++) {
            char* item = ((char**)result.data)[i];
            PyObject* py_item = item==NULL ? GetPyNone() : PyUnicode_FromString(item);
            free(item);
            PyList_SetItem(py_result, i, py_item);
        }
    }
    free(result.data);
    return py_result;
}"""
test_cases["return_string_slice"] = (fn, fn_res)


fn = GoFunction(
    name="Func_5",
    arguments=[],
    return_type=[{"go_type": "[][]byte"}, {"go_type": "[]*C.char"}],
)
fn_res = """
static PyObject* test_func_5(PyObject* self, PyObject* args) { 
    struct Func_5_return result = Func_5();
    PyObject* py_result_r0;
    if (result.r0.data == NULL) {
        py_result_r0 = GetPyNone();
    } else {
        py_result_r0 = PyList_New(result.r0.len);
        for (int i = 0; i < result.r0.len; i++) {
            GoSlice item = ((GoSlice*)result.r0.data)[i];
            PyObject* py_item = item.data==NULL ? GetPyNone() : PyBytes_FromStringAndSize(item.data, item.len);
            free(item.data);
            PyList_SetItem(py_result_r0, i, py_item);
        }
    }
    free(result.r0.data);
    PyObject* py_result_r1;
    if (result.r1.data == NULL) {
        py_result_r1 = GetPyNone();
    } else {
        py_result_r1 = PyList_New(result.r1.len);
        for (int i = 0; i < result.r1.len; i++) {
            char* item = ((char**)result.r1.data)[i];
            PyObject* py_item = item==NULL ? GetPyNone() : PyUnicode_FromString(item);
            free(item);
            PyList_SetItem(py_result_r1, i, py_item);
        }
    }
    free(result.r1.data);
    PyObject* py_result = Py_BuildValue("OO", py_result_r0, py_result_r1);
    Py_DECREF(py_result_r0);
    Py_DECREF(py_result_r1);
    return py_result;
}"""
test_cases["return_string&byte_slices"] = (fn, fn_res)


@pytest.mark.parametrize("fn,fn_res", test_cases.values(), ids=test_cases.keys())
def test_gen_fn(fn, fn_res):
    assert gen_fn(fn, "test") == fn_res


"""
uv run -m tests.test_complex_functions
"""



from pydantic import BaseModel
from go4py.code_gen.file_gen import gen_binding_file
from go4py.get_go_functions import get_go_functions
import argparse
import os
import yaml

from go4py.types import go4pyConfig


def parse_args():
    parser = argparse.ArgumentParser(description="Generate Go binding file")
    parser.add_argument("module_path", help="Name (or Path) of the module")
    return parser.parse_args()


def read_config():
    yaml_file = "go4py.yaml"
    if os.path.exists(yaml_file):
        with open(yaml_file, "r") as file:
            go4py_config = yaml.safe_load(file)
            return go4py_config
    return {}


if __name__ == "__main__":
    args = parse_args()
    module_path = args.module_path
    functions = get_go_functions(module_path)
    module_name = module_path.split("/")[-1]

    config = go4pyConfig.model_validate(read_config())
    config.module_name = module_name

    gen_binding_file(config, functions, "cpython-extention/autogen_bindings.c")



from pydantic import BaseModel
from typing import ClassVar, Literal, TypeAlias

from abc import ABC, abstractmethod
import logging

from go4py.doc_annotation import DocAnnots, make_doc_annots

logger = logging.getLogger(__name__)


class go4pyConfig(BaseModel):
    custom_incudes: list[str] = []
    custom_methods: list[str] = []
    output_file: str = "cpython-extention/bindings.c"
    module_name: str = ""


# cusotom exception: cgo limitation
class CgoLimitationError(Exception):
    pass


# go_types_dict = {
#     # here we assume 64-bit system
#     "int": IntType(bits=64),
#     "int8": IntType(bits=8),
#     "int16": IntType(bits=16),
#     "int32": IntType(bits=32),
#     "int64": IntType(bits=64),
#     "uint": IntType(bits=64, unsigned=True),
#     "uint8": IntType(bits=8, unsigned=True),
#     "uint16": IntType(bits=16, unsigned=True),
#     "uint32": IntType(bits=32, unsigned=True),
#     "uint64": IntType(bits=64, unsigned=True),
#     "byte": IntType(bits=8, unsigned=True),
#     "rune": IntType(bits=32),
#     "float32": FloatType(bits=32),
#     "float64": FloatType(bits=64),
#     "bool": BoolType(),
#     "string": StringType(),
#     # "error": ErrorType(),
#     # uintptr": None,
# }


class VarType(BaseModel, ABC):
    go_type: str
    need_copy: ClassVar[bool]

    @abstractmethod
    def c_type(self) -> str: ...

    @abstractmethod
    def fmt_str(self) -> str: ...

    @abstractmethod
    def converter(self, inp: str): ...

    @abstractmethod
    def from_py_converter(self, inp: str): ...

    @abstractmethod
    def check(self, inp): ...

    def need_free(self):
        return False

    def cgo_type(self) -> str:
        return self.c_type()


class IntType(VarType):
    go_type: Literal[
        "int",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "byte",
        "rune",
    ] = "int"
    bits: int = 64
    unsigned: bool = False
    need_copy: ClassVar[bool] = False

    def model_post_init(self, __context):
        if self.go_type in ["int", "uint"]:
            self.bits = 64
        elif self.go_type in ["int8", "uint8", "byte"]:
            self.bits = 8
        elif self.go_type in ["int16", "uint16"]:
            self.bits = 16
        elif self.go_type in ["int32", "uint32", "rune"]:
            self.bits = 32
        elif self.go_type in ["int64", "uint64"]:
            self.bits = 64
        else:
            raise ValueError(f"Unsupported go type: {self.go_type}")
        if self.go_type in ["uint", "uint8", "uint16", "uint32", "uint64", "byte"]:
            self.unsigned = True
        else:
            self.unsigned = False

    def c_type(self) -> str:
        _sign = "" if not self.unsigned else "unsigned "
        if self.bits == 8:
            return _sign + "char"
        elif self.bits == 16:
            return _sign + "short"
        elif self.bits == 32:
            return _sign + "int"
        elif self.bits == 64:
            return _sign + "long"
        else:
            raise ValueError(f"Unsupported bit size: {self.bits}")

    def fmt_str(self) -> str:
        """Returns the format string for the given integer type."""
        if self.bits == 8:
            return "b" if not self.unsigned else "B"
        elif self.bits == 16:
            return "h" if not self.unsigned else "H"
        elif self.bits == 32:
            return "i" if not self.unsigned else "I"
        elif self.bits == 64:
            return "l" if not self.unsigned else "L"
        else:
            raise ValueError(f"Unsupported bit size: {self.bits}")

    def converter(self, inp):
        if self.unsigned:
            return f"PyLong_FromUnsignedLong({inp})"
        else:
            return f"PyLong_FromLong({inp})"

    def from_py_converter(self, inp):
        if self.bits == 64:
            if self.unsigned:
                return f"PyLong_AsUnsignedLong({inp})"
            else:
                return f"PyLong_AsLong({inp})"
        elif self.bits == 32:
            if self.unsigned:
                raise Exception(
                    "no python function found for conversion to uint32 use int64, uint64 or int32 instead"
                )
            else:
                return f"PyLong_AsInt({inp})"
        else:
            raise Exception(
                f"no python function found for conversion to {self.c_type()} use int64, uint64 or int32 instead"
            )

    def check(self, inp):
        return f"PyLong_Check({inp})"


class FloatType(VarType):
    go_type: Literal["float64", "float32"] = "float"
    bits: int = 64
    need_copy: ClassVar[bool] = False

    def model_post_init(self, __context):
        if self.go_type == "float32":
            self.bits = 32
        elif self.go_type == "float64":
            self.bits = 64
        else:
            raise ValueError(f"Unsupported float type: {self.go_type}")

    def c_type(self) -> str:
        if self.bits == 32:
            return "float"
        elif self.bits == 64:
            return "double"
        else:
            raise ValueError(f"Unsupported bit size: {self.bits}")

    def fmt_str(self) -> str:
        if self.bits == 32:
            return "f"
        elif self.bits == 64:
            return "d"
        else:
            raise ValueError(f"Unsupported bit size: {self.bits}")

    def converter(self, inp):
        if self.bits == 32:
            return f"PyFloat_FromFloat({inp})"
        elif self.bits == 64:
            return f"PyFloat_FromDouble({inp})"
        else:
            raise ValueError(f"Unsupported bit size: {self.bits}")

    def from_py_converter(self, inp):
        if self.bits == 64:
            return f"PyFloat_AsDouble({inp})"
        else:
            logger.warning(f"casting double to float in C for {inp} (may overflow silently)")
            return f"(float)PyFloat_AsDouble({inp})"

    def check(self, inp):
        return f"PyFloat_Check({inp})"


class BoolType(VarType):
    go_type: Literal["bool"] = "bool"
    need_copy: ClassVar[bool] = False

    def c_type(self) -> str:
        return "int"

    def fmt_str(self) -> str:
        return "p"

    def converter(self, inp):
        return f"PyBool_FromLong({inp})"

    def from_py_converter(self, inp):
        return f"PyObject_IsTrue({inp})"

    def check(self, inp):
        return None  # skip checking for bool


class CStringType(VarType):
    go_type: Literal["*C.char"] = "*C.char"
    need_copy: ClassVar[bool] = False

    def c_type(self) -> str:
        return "char*"

    def fmt_str(self) -> str:
        return "s"

    def converter(self, inp):
        return f"PyUnicode_FromString({inp})"

    def from_py_converter(self, inp):
        return f"PyUnicode_AsUTF8({inp})"

    def check(self, inp):
        return f"PyUnicode_Check({inp})"

    def need_free(self):
        return True

    def cgo_type(self):
        return "char*"


class GoStringType(VarType):
    go_type: Literal["string"] = "string"
    need_copy: ClassVar[bool] = True

    def c_type(self) -> str:
        return "char*"

    def fmt_str(self) -> str:
        return "s"

    def converter(self, inp):
        raise CgoLimitationError("Don't return string from cgo")

    def from_py_converter(self, inp):
        return f"PyUnicode_AsUTF8({inp})"

    def check(self, inp):
        return f"PyUnicode_Check({inp})"

    def need_free(self):
        return True

    def cgo_type(self):
        return "GoString"


class ByteSliceType(VarType):
    go_type: Literal["[]byte"] = "[]byte"
    need_copy: ClassVar[bool] = True
    size: str | None = None

    def c_type(self) -> str:
        return "PyObject*"

    def fmt_str(self) -> str:
        return "O"

    def converter(self, inp):
        return f"PyBytes_FromStringAndSize({inp}.data, {inp}.len)"

    def from_py_converter(self, inp):
        return NotImplementedError()

    def need_free(self):
        return True

    def check(self, inp):
        return f"PyBytes_Check({inp})"

    def cgo_type(self):
        return "GoSlice"


class UnknownType(VarType):
    go_type: str
    need_copy: bool = True

    def c_type(self):
        raise NotImplementedError()

    def fmt_str(self): ...

    def converter(self, inp): ...

    def from_py_converter(self, inp): ...

    def check(self, inp): ...

    def need_free(self): ...

    def resolve(self):
        """try to convert this type to other types"""
        if self.go_type.startswith("[]"):
            t = self.go_type[2:]
            if t.startswith("(") and t.endswith(")"):
                t = t[1:-1]
            return SliceType(item_type={"go_type": t})
        else:
            raise CgoLimitationError(f"please check if you can use type: {self.go_type}")


SimpleTypes: TypeAlias = (
    IntType
    | FloatType
    | BoolType
    | GoStringType
    | CStringType
    | ByteSliceType
    | UnknownType  # Complex Types will fall into this but will be resolved later
)


class SliceType(VarType):
    go_type: Literal["Slice"] = "Slice"
    item_type: SimpleTypes
    need_copy: ClassVar[bool] = True

    def c_type(self) -> str:
        return "PyObject*"

    def fmt_str(self) -> str:
        return "O"

    def converter(self, inp):
        raise NotImplementedError()

    def from_py_converter(self, inp):
        return NotImplementedError()

    def check(self, inp):
        return f"PyList_Check({inp})"

    def need_free(self):
        return True

    def cgo_type(self) -> str:
        return "GoSlice"


RealType: TypeAlias = SimpleTypes | SliceType


class Variable(BaseModel):
    name: str | None
    type: RealType


class GoFunction(BaseModel):
    name: str
    arguments: list[Variable]
    return_type: list[RealType]
    package: str = ""
    docs: str = ""

    def __str__(self) -> str:
        return f"{self.package}.{self.name}"

    def lowercase_name(self) -> str:
        return self.name.lower()[0] + self.name[1:]

    def doc_annots(self) -> DocAnnots:
        return make_doc_annots(self.docs)



import pytest
from go4py.code_gen.generate_wrapper import gen_fn
from go4py.types import GoFunction

test_cases = {}

fn = GoFunction(
    name="Func_12",
    docs="[go4py] msgpack-decode\n",
    arguments=[],
    return_type=[{"go_type": "[]byte"}, {"go_type": "*C.char"}],
)
fn_res = """
static PyObject* test_func_12(PyObject* self, PyObject* args) { 
    struct Func_12_return result = Func_12();
    PyObject* py_result_r0_msgpack;
    if (result.r0.data!=NULL){
        PyObject* py_result_r0 = PyBytes_FromStringAndSize(result.r0.data, result.r0.len);
        py_result_r0_msgpack = PyObject_CallFunctionObjArgs(unpackb, py_result_r0, NULL);
        Py_DECREF(py_result_r0);
    }else{
        py_result_r0_msgpack = GetPyNone();
    }
    free(result.r0.data);
    PyObject* py_result_r1 = result.r1==NULL ? GetPyNone() : PyUnicode_FromString(result.r1);
    free(result.r1);
    PyObject* py_result = Py_BuildValue("OO", py_result_r0_msgpack, py_result_r1);
    Py_DECREF(py_result_r0_msgpack);
    Py_DECREF(py_result_r1);
    return py_result;
}"""
test_cases["msgpack_decode_return"] = (fn, fn_res)


fn = GoFunction(
    name="Func_13",
    docs="[go4py] skip-binding\n",
    arguments=[],
    return_type=[{"go_type": "[]byte"}, {"go_type": "*C.char"}],
)
fn_res = "\n// function Func_13 is skipped due to 'skip-binding' annotation\n"
test_cases["skip-binding"] = (fn, fn_res)

fn = GoFunction(
    name="Func_14",
    docs="[go4py] no-gil\n",
    arguments=[{"name": "arg0", "type": {"go_type": "int"}}],
    return_type=[{"go_type": "[]byte"}],
)
fn_res = """
static PyObject* test_func_14(PyObject* self, PyObject* args) { 
    long arg0;
    if (!PyArg_ParseTuple(args, "l", &arg0))
        return NULL;
    GoSlice result;
    Py_BEGIN_ALLOW_THREADS
    result = Func_14(arg0);
    Py_END_ALLOW_THREADS
    PyObject* py_result = result.data==NULL ? GetPyNone() : PyBytes_FromStringAndSize(result.data, result.len);
    free(result.data);
    return py_result;
}"""
test_cases["no_gil_function"] = (fn, fn_res)

print(gen_fn(fn, "test"))


@pytest.mark.parametrize("fn,fn_res", test_cases.values(), ids=test_cases.keys())
def test_gen_fn(fn, fn_res):
    assert gen_fn(fn, "test") == fn_res


"""
uv run -m tests.test_annotations
"""



import pytest

from go4py.code_gen.generate_wrapper import gen_fn
from go4py.types import GoFunction


test_cases = {}

fn = GoFunction(
    name="Func_1",
    arguments=[],
    return_type=[{"go_type": "int"}],
)
fn_res = """
static PyObject* test_func_1(PyObject* self, PyObject* args) { 
    long result = Func_1();
    return PyLong_FromLong(result);
}"""
test_cases["simple_int_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_two",
    arguments=[],
    return_type=[{"go_type": "int"}, {"go_type": "float32"}],
)
fn_res = """
static PyObject* test_func_two(PyObject* self, PyObject* args) { 
    struct Func_two_return result = Func_two();
    PyObject* py_result = Py_BuildValue("lf", result.r0, result.r1);
    return py_result;
}"""
test_cases["tuple_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_3",
    arguments=[{"name": "a", "type": {"go_type": "int8"}}],
    return_type=[],
)

fn_res = """
static PyObject* test_func_3(PyObject* self, PyObject* args) { 
    char a;
    if (!PyArg_ParseTuple(args, "b", &a))
        return NULL;
    Func_3(a);
    RETURN_NONE;
}"""
test_cases["int8_arg"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_4",
    arguments=[{"name": "a", "type": {"go_type": "string"}}],
    return_type=[],
)

fn_res = """
static PyObject* test_func_4(PyObject* self, PyObject* args) { 
    char* a;
    if (!PyArg_ParseTuple(args, "s", &a))
        return NULL;
    GoString go_a = {a, (GoInt)strlen(a)};
    Func_4(go_a);
    RETURN_NONE;
}"""
test_cases["string_arg"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_x",
    arguments=[],
    return_type=[{"go_type": "*C.char"}],
)
fn_res = """
static PyObject* test_func_x(PyObject* self, PyObject* args) { 
    char* result = Func_x();
    PyObject* py_result = result==NULL ? GetPyNone() : PyUnicode_FromString(result);
    free(result);
    return py_result;
}"""
test_cases["str_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_4x",
    arguments=[],
    return_type=[{"go_type": "*C.char"}, {"go_type": "[]byte"}],
)
fn_res = """
static PyObject* test_func_4x(PyObject* self, PyObject* args) { 
    struct Func_4x_return result = Func_4x();
    PyObject* py_result_r0 = result.r0==NULL ? GetPyNone() : PyUnicode_FromString(result.r0);
    free(result.r0);
    PyObject* py_result_r1 = result.r1.data==NULL ? GetPyNone() : PyBytes_FromStringAndSize(result.r1.data, result.r1.len);
    free(result.r1.data);
    PyObject* py_result = Py_BuildValue("OO", py_result_r0, py_result_r1);
    Py_DECREF(py_result_r0);
    Py_DECREF(py_result_r1);
    return py_result;
}"""
test_cases["str&bytes_tuple_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_5",
    arguments=[],
    return_type=[{"go_type": "*C.char"}, {"go_type": "*C.char"}],
)
fn_res = """
static PyObject* test_func_5(PyObject* self, PyObject* args) { 
    struct Func_5_return result = Func_5();
    PyObject* py_result_r0 = result.r0==NULL ? GetPyNone() : PyUnicode_FromString(result.r0);
    free(result.r0);
    PyObject* py_result_r1 = result.r1==NULL ? GetPyNone() : PyUnicode_FromString(result.r1);
    free(result.r1);
    PyObject* py_result = Py_BuildValue("OO", py_result_r0, py_result_r1);
    Py_DECREF(py_result_r0);
    Py_DECREF(py_result_r1);
    return py_result;
}"""
test_cases["str_tuple_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_6",
    arguments=[{"name": "a", "type": {"go_type": "string"}}],
    return_type=[{"go_type": "*C.char"}],
)
fn_res = """
static PyObject* test_func_6(PyObject* self, PyObject* args) { 
    char* a;
    if (!PyArg_ParseTuple(args, "s", &a))
        return NULL;
    GoString go_a = {a, (GoInt)strlen(a)};
    char* result = Func_6(go_a);
    PyObject* py_result = result==NULL ? GetPyNone() : PyUnicode_FromString(result);
    free(result);
    return py_result;
}"""
test_cases["string_arg_string_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_7",
    arguments=[
        {"name": "a", "type": {"go_type": "string"}},
        {"name": "b", "type": {"go_type": "uint16"}},
        {"name": "c", "type": {"go_type": "string"}},
        {"name": "d", "type": {"go_type": "float64"}},
    ],
    return_type=[{"go_type": "bool"}],
)
fn_res = """
static PyObject* test_func_7(PyObject* self, PyObject* args) { 
    char* a;
    unsigned short b;
    char* c;
    double d;
    if (!PyArg_ParseTuple(args, "sHsd", &a, &b, &c, &d))
        return NULL;
    GoString go_a = {a, (GoInt)strlen(a)};
    GoString go_c = {c, (GoInt)strlen(c)};
    int result = Func_7(go_a,b,go_c,d);
    return PyBool_FromLong(result);
}"""
test_cases["multi_arg_bool_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_8",
    arguments=[
        {"name": "a", "type": {"go_type": "*C.char"}},
    ],
    return_type=[],
)
fn_res = """
static PyObject* test_func_8(PyObject* self, PyObject* args) { 
    char* a;
    if (!PyArg_ParseTuple(args, "s", &a))
        return NULL;
    Func_8(a);
    RETURN_NONE;
}"""
test_cases["char_ptr_no_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_9",
    arguments=[
        {"name": "a", "type": {"go_type": "[]byte"}},
    ],
    return_type=[],
)
fn_res = """
static PyObject* test_func_9(PyObject* self, PyObject* args) { 
    PyObject* a;
    if (!PyArg_ParseTuple(args, "O", &a))
        return NULL;
    GoInt len_a = PyBytes_Size(a);
    GoSlice go_a = {PyBytes_AsString(a), len_a, len_a};
    Func_9(go_a);
    RETURN_NONE;
}"""
test_cases["bytes_input"] = (fn, fn_res)

######

fn = GoFunction(
    name="Func_10",
    arguments=[],
    return_type=[{"go_type": "[]byte"}],
)
fn_res = """
static PyObject* test_func_10(PyObject* self, PyObject* args) { 
    GoSlice result = Func_10();
    PyObject* py_result = result.data==NULL ? GetPyNone() : PyBytes_FromStringAndSize(result.data, result.len);
    free(result.data);
    return py_result;
}"""
test_cases["bytes_return"] = (fn, fn_res)

#####

fn = GoFunction(
    name="Func_11",
    arguments=[],
    return_type=[{"go_type": "[]byte"}, {"go_type": "[]byte"}],
)
fn_res = """
static PyObject* test_func_11(PyObject* self, PyObject* args) { 
    struct Func_11_return result = Func_11();
    PyObject* py_result_r0 = result.r0.data==NULL ? GetPyNone() : PyBytes_FromStringAndSize(result.r0.data, result.r0.len);
    free(result.r0.data);
    PyObject* py_result_r1 = result.r1.data==NULL ? GetPyNone() : PyBytes_FromStringAndSize(result.r1.data, result.r1.len);
    free(result.r1.data);
    PyObject* py_result = Py_BuildValue("OO", py_result_r0, py_result_r1);
    Py_DECREF(py_result_r0);
    Py_DECREF(py_result_r1);
    return py_result;
}"""
test_cases["byte_tuple_return"] = (fn, fn_res)


@pytest.mark.parametrize("fn,fn_res", test_cases.values(), ids=test_cases.keys())
def test_gen_fn(fn, fn_res):
    assert gen_fn(fn, "test") == fn_res


"""
uv run -m tests.test_simple_functions
"""



import os
from pathlib import Path
from go4py.code_gen.generate_wrapper import gen_fn
from go4py.types import CgoLimitationError, GoFunction, go4pyConfig


def template(config: go4pyConfig, functions_code: list, methods: str):
    custom_incudes = "\n".join(config.custom_incudes)
    custom_methods = "".join(["\n    " + m for m in config.custom_methods])

    return f"""
#define PY_SSIZE_T_CLEAN

#include <Python.h>
#include <string.h>
#include "../artifacts/build/lib{config.module_name}.h"
{custom_incudes}

#define RETURN_NONE Py_INCREF(Py_None) ; return Py_None
PyObject* GetPyNone() {{
    Py_INCREF(Py_None);
    return Py_None;
}}

PyObject* unpackb;
{functions_code}

static PyMethodDef Methods[] = {{{methods}{custom_methods}
    {{NULL, NULL, 0, NULL}}
}};

static struct PyModuleDef {config.module_name}_module = {{
    PyModuleDef_HEAD_INIT,
    "{config.module_name}",
    NULL,
    -1,
    Methods
}};
PyMODINIT_FUNC PyInit_{config.module_name}(void) {{
    PyObject* msgpack = PyImport_ImportModule("msgpack");
    if (msgpack == NULL) {{
       PyErr_SetString(PyExc_ImportError, "msgpack module not found");
        return NULL;
    }}
    unpackb = PyObject_GetAttrString(msgpack, "unpackb");

    return PyModule_Create(&{config.module_name}_module);
}}
"""


def gen_binding_file(config: go4pyConfig, functions: list[GoFunction], dest: Path | str):
    module = config.module_name
    functions_code = ""
    res_functions: list[GoFunction] = []
    for fn in functions:
        try:
            functions_code += "\n" + gen_fn(fn, module)
            res_functions.append(fn)
        except CgoLimitationError as e:
            print("[cgo limitation]", e, "-> skipping function: ", fn.name)
            continue
        except Exception as e:
            print("[error]", e, "-> skipping function: ", fn.name)
            continue

    methods = ""
    for fn in res_functions:
        if fn.doc_annots().skip_binding:
            continue
        fn_name = fn.lowercase_name()
        methods += f'\n    {{"{fn_name}", {module}_{fn_name}, METH_VARARGS, "{fn_name}"}},'
    os.makedirs(os.path.dirname(dest), exist_ok=True)
    with open(dest, "w") as f:
        f.write(template(config, functions_code, methods))



import sys
import click
import os
import subprocess
from pathlib import Path

from go4py.template_engine import render_template
from go4py.utils.text_util import print_text_box

HERE = Path(__file__).parent
TEMPLATE_DIR = HERE / "templates"


def find_make_files(build_path):
    build_path = Path(build_path)
    if not build_path.exists():
        click.echo(f"Error: Package directory '{build_path}' does not exist.")
        sys.exit(1)

    makefiles = ("Makefile", "makefile")
    if any((build_path / f).exists() for f in makefiles):
        yield build_path
    else:
        for item in build_path.iterdir():
            if item.is_dir():
                if any((item / f).exists() for f in makefiles):
                    yield item


@click.group()
def cli():
    """go4py CLI"""
    pass


@cli.command()
@click.argument("module_name")
def init(module_name):
    """Initialize a go module with the given name."""

    module_dir = Path(module_name)
    if module_dir.exists():
        click.echo(f"A directory with the name of {module_name} already exists.")
        return
    # create directory with the name of the module and cd there
    module_dir.mkdir()

    # Check if we're already inside a Go module (check current and parent directory)
    current_go_mod_exists = Path("go.mod").exists()
    parent_go_mod_exists = Path("..").joinpath("go.mod").exists()

    if not (current_go_mod_exists or parent_go_mod_exists):
        # initialize a go module in root directory
        os.system(f"go mod init {module_name}")
    else:
        click.echo("Already inside a Go module, skipping 'go mod init'")

    # for all files in template folder copy them to the module folder
    data = {"module_name": module_name.split("/")[-1]}
    for file in TEMPLATE_DIR.rglob("*"):
        if file.is_file():
            dst_file = module_dir / file.relative_to(TEMPLATE_DIR)
            dst_file.parent.mkdir(parents=True, exist_ok=True)
            dst_file.write_text(render_template(file.read_text(), data))

    # print a message
    click.echo(f"Module {module_name} initialized.")


@cli.command()
@click.argument("build_path", required=False, default=".")
def build(build_path):
    """Build a go module by running make in its directory.

    If module_name is not provided, searches all directories for Makefiles
    and runs make in each directory that has one.
    """
    build_path = Path(build_path)
    if not build_path.exists():
        click.echo(f"Error: Package directory '{build_path}' does not exist.")
        exit(1)

    for makepath in find_make_files(build_path):
        click.echo(f"Building module '{makepath}'...")
        subprocess.run(["make", "-C", str(makepath), "-j2"])


@cli.command()
@click.argument("build_path", required=False, default=".")
def clean(build_path):
    """Clean all build artifacts."""
    for makepath in find_make_files(build_path):
        subprocess.run(["make", "-C", str(makepath), "clean"])


@cli.command()
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def parse(args):
    """parse go files in a directory to find exported functions."""
    exetutable = HERE / "parse"
    if not exetutable.exists():
        # try to build it
        os.system(f"go build -o {str(exetutable)} {HERE / './go_cmd/parse'}")
    if not exetutable.exists():
        click.echo("Error: parse executable not found.")
        # exit with error
        exit(1)
    cmd = [str(exetutable)] + list(args)
    subprocess.run(cmd)


@cli.command()
@click.argument("text")
@click.argument("bold_text", default="")
def textbox(text, bold_text):
    print_text_box(text, bold_text)


if __name__ == "__main__":
    cli()



from pydantic import BaseModel


class DocAnnots(BaseModel):
    msgpack_decode: bool = False
    no_gil: bool = False
    skip_binding: bool = False


def make_doc_annots(doc: str) -> DocAnnots:
    doc_annots = DocAnnots()
    for line in doc.splitlines():
        if "[go4py]" in line:
            args = line.replace("[go4py]", "").strip().split()
            for arg in args:
                if arg == "msgpack-decode":
                    doc_annots.msgpack_decode = True
                elif arg == "no-gil":
                    doc_annots.no_gil = True
                elif arg == "skip-binding":
                    doc_annots.skip_binding = True
                else:
                    raise ValueError(f"Unknown go4py annotation: {arg}")
    return doc_annots


def test_make_doc_annots():
    doc = "[go4py]   msgpack-decode\n"
    doc_annots = make_doc_annots(doc)
    assert doc_annots.msgpack_decode


if __name__ == "__main__":
    test_make_doc_annots()
    print("All tests passed.")

"""
uv run -m go4py.doc_annotation
"""
